///////////////////////////////////////////////////////////////////////////////
//
//                             A  C  S  N
//
// +---------------  Alaska Certified Solutions Network  -------------------+
// |                                                                        |
// |        This file is proved and certified by Alaska Software            |
// |                                                                        |
// |                   No: <Certification number>                           |
// |                          109xxx-04-0014                                |
// |                                                                        |
// |   For more information about ACSN read the appropriate announcement    |
// |      or scan for ACSN in the Alaska Support-LIBs on CompuServe or      |
// |                   at WWW.ALASKA-SOFTWARE.COM                           |
// |                                                                        |
// +------------------------------------------------------------------------+
//
// FILE NAME
//
//    Eventspy.prg
//
// AUTHOR
//
//    (c) Copyright 1998-99, Hannes Ziegler
//
//    ALL RIGHTS RESERVED
//
//    This file is the property of AUTHOR. It participates in the
//    Alaska Certified Solutions Network program. Permission to use,
//    copy, modify, and distribute this software for any purpose and
//    without fee is hereby granted, provided that the above copyright
//    notice appear in all copies and that the name of the author or
//    Alaska Software not be used in advertising or publicity pertaining
//    to distribution of the software without specific, written prior
//    permission.
//
// WARRANTY
//
//    THE MATERIAL EMBODIED ON THIS SOFTWARE IS PROVIDED TO YOU "AS-IS"
//    AND WITHOUT WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR OTHERWISE,
//    INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF MERCHANTABILITY OR
//    FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL THE AUTHOR
//    OR ALASKA SOFTWARE BE LIABLE TO YOU OR ANYONE ELSE FOR ANY DIRECT,
//    SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
//    INCLUDING WITHOUT LIMITATION, LOSS OF PROFIT AND LOSS OF USE.
//
// DESCRIPTION
//
//    This file implements an Event Spy that can display all events,
//    message parameters and message receivers in an Xbase++ program.
//
//    It is great for the Xbase++ novice who wants to see what happens
//    in the event loop, and it is a great tool for advanced programmers
//    who need to trace single events.
//
//    FYI: I have found this tool to be invaluable for "hard core" event
//    programming. That is: a single Xbase Part captures all mouse
//    messages. In this scenario, it is pretty hard to find a paint event
//    that causes wrong display unless you have a complete history of
//    events and their receivers.
//
// USAGE
//
//    Copy EVENTSPY.DLL and EVENTSPY.LIB to a directory of the LIB path.
//    (e.g: \ALASKA\XPPW32\LIB). And link EVENTSPY.LIB to your application
//    using the link switch /PM:PM or set GUI=YES in your project file
//
// WHAT'S NEW IN THIS VERSION
//
//    * There is no need to change your code with #commands like
//      INIT EVENTSPY or SPY anymore. This was required in the first 
//      Spy release. With the new version, you can just link EVENTSPY.LIB
//      to have the Event Spy available in your application. 
//
//    * APPEVENT.CH is parsed only if APPEVENT.SPY is not found in the
//      include directory.
//      This shortens the load time of the Event Spy considerably.
//      If APPEVENT.CH changes, delete APPEVENT.SPY (it will be rebuilt)
//      or re-load it from the Options window.
//      
//    * Replaced PUBLIC variable 'EventSpy' with Function EventSpy()
//    
//    * The Event Spy can be configured now for optimal usage.
//      I have received great ideas from people using the Event Spy and
//      have tried to incorporate all the goodies that make this tool
//      more useful (special thanks go to Scott McNay and Ken Levitt).
//
//    * Configuration data is saved in OPTIONS.SPY so that the Event Spy
//      "wakes up" in the same state as it is left. The config file is
//      saved in the directory where EVENTSPY.DLL is located.
//    
//    * Configuration options for the Spy are (implemented in OPTIONS.PRG)
//    
//      - stay on top
//      - add events when Spy is hidden
//      - use fixed font
//      - auto-adjust to Log width
//      - mark XBP having focus in the log
//      - show thread ID
//      - show time stamp
//      - display short event constants
//      - show/ignore 2nd message param
//      - filter events sent to Spy
//      - filter Paint events
//      - filter Motion Events
//      - define hot key for restoring hidden Spy
//      - define max. number of Log entries
//
//      Note that raw event data is stored in the Event Spy so that
//      external routines can access the event log. However, this has
//      the drawback that Xbase Parts are referenced in the Spy and
//      are not removed from memory after :destroy() until the event log
//      is cleared or a log entry is removed due to the max. # setting.
//
//    * Features for use by external routines are available that let
//      you look up/investigate events and the event log:
//
//      - Callback code block when a log entry is double clicked
//          EventSpy():logSelected := {|aEvent| YourFunc( aEvent ) }
//
//          aEvent is a 4 element array { nEvent, mp1, mp2, oXbp }
//          that is passed to YourFunc()
//
//      - Retrieve raw event data from the Log (e.g. 10th last event)
//          EventSpy():getLogEntry( 10 )   --> aEvent 
// 
//      - Transfer APPEVENT.CH data into your program
//          EventSpy():readAppEventCH( @aDefines, @aValues )
//
//      - Translate a char event constant to numeric and vice versa
//          EventSpy():translate( nEvent  ) --> cDefine
//          EventSpy():translate( cDefine ) --> nEvent
//
//      - Save the current event Log to EVENTSPY.LOG
//          EventSpy():saveLog()
//
// REMARKS
//
//    The textual information for the numeric events comes from the
//    APPEVENT.CH file. The program searches for this file in the
//    SET INCLUDE= path and parses #define constants and their numeric
//    values from this file. If APPEVENT.CH is not found, the program
//    simply QUITs.
//
//    The Event Spy window is fully resizable. Note the variable distance
//    between the pushbuttons when you change the width
//    of the window. This is accomplished by the GroupManager class.
//
//    Open the Options Window and check all checkboxes in the Settings,
//    Info and Filter tab pages for initial testing.
//
// HISTORY
//
//    02/12/99 Added auto-adjustment to log-width, input focus mark 
//                   and display of short event constants
//                   (credit to Ken Levitt)
//                   
//    02/08/99 Added fixed font usage to enhance readability of the
//                   log entries.
//                 
//    02/06/99 Merged Scott McNay's great contributions and his bug fixes.
//                   These are in particular:
//                 - force Spy window to front
//                 - force log to be updated even when Spy is hidden
//                 - time stamp to log entries.  Sometimes this is
//                   helpful for figuring things out after the fact.
//                 - saving log to file EVENTSPY.LOG
//                 - Exported :ReadAppEventCH() so that other programs can
//                   access the AppEvent.CH data.  
//                 - Fix for event loops that have a timeout.
//                 - Fix ToString to validate datatype of :getTitle()
//                 - Raw event info is now available so that the
//                   user can select a specific event for closer review,
//                   and/or so the other program itself can directly look
//                   up prior events.
//
//          HZ Note: I did not use Scott's original code since some
//                   of his ideas went into the new Options window and
//                   for other ideas I have chosen a different implementation.
//                   However, if you find these enhancements useful (I do),
//                   your credit goes to Scott McNay.
//
//    01/31/99 Added an Options window that allows for configuring the
//                   EventSpy. The check boxes went into that window
//                   Configuration data is stored in a file and re-used 
//                   in a next Spy session.
//
//    08/08/98 Added binary storage of data parsed from APPEVENT.CH so that
//                   the EventSpy is immediately visible at program start
//             Added function mapping APPEVENT = SPYAPPEVENT and removed
//                   EVENTSPY.CH. It's more convenient to just link
//                   EVENTSPY.LIB to the EXE than using #commands, even if
//                   they are quickly typed.
//
//    07/02/98 Added EVENTSPY.CH to the project and moved the Spy into a DLL.
//    06/26/98 Front-end revised, Comments and ACSN header added.
//    05/12/98 Front-end finished.
//    03/04/98 Parsing of APPEVENT.CH finished.
//
///////////////////////////////////////////////////////////////////////////////

#include "Appevent.ch"
#include "Common.ch"
#include "Fileio.ch"
#include "Font.ch"
#include "Gra.ch"
#include "Xbp.ch"


#define SPY_EVENT_FILE     "APPEVENT.SPY"  // Binary Appevent.ch data file

#define CRLF               Chr(13)+Chr(10)


STATIC soSpy

******************************************************************************
* This procedure creates the Event Spy window at program start
******************************************************************************
INIT PROCEDURE CreateEventSpy
   LOCAL nEvent, mp1, mp2, oXbp

   #pragma Info( NOINIT )
   IF soSpy == NIL
      soSpy := EventSpyDialog():New():Create() 
      soSpy:show()
      soSpy:setFrameState( soSpy:options:frameState )

      // Process events until spy window is completely drawn
      DO WHILE NextAppEvent() <> 0
         nEvent := Appevent( @mp1, @mp2, @oXbp )
         oXbp:handleEvent( nEvent, mp1, mp2 )
      ENDDO
   ENDIF
   #pragma Info( INIT )
RETURN


******************************************************************************
* This procedure saves the current Event Spy options at program termination
* Note: terminating the EXE with ALT-C does not save the options
******************************************************************************
EXIT PROCEDURE SaveEventSpyOptions
   EventSpyOptions():Save( soSpy:options )
RETURN


******************************************************************************
* This funtion returns a reference to the Event Spy window
******************************************************************************
FUNCTION EventSpy
RETURN soSpy


******************************************************************************
* This function is called when EVENTSPY.LIB is linked to the EXE
* for each call to AppEvent() occurring in the EXE
* This is due to the APPEVENT = SPYAPPEVENT mapping in the DEF file
******************************************************************************
FUNCTION SpyAppevent( mp1, mp2, oXbp, nTimeOut )
   LOCAL nEvent := Appevent( @mp1, @mp2, @oXbp, nTimeOut )

   soSpy:addEvent( nEvent, mp1, mp2, oXbp )

   IF nEvent == xbeP_Keyboard .AND. mp1 == soSpy:options:hotKey
      IF soSpy:getFrameState() == XBPDLG_FRAMESTAT_MINIMIZED 
         soSpy:setFrameState( XBPDLG_FRAMESTAT_NORMALIZED )
      ENDIF
      soSpy:show()
      SetAppFocus( soSpy )
   ENDIF
RETURN nEvent


******************************************************************************
* A class that displays events in readable form using XbpTreeView
******************************************************************************
STATIC CLASS EventSpyDialog FROM XbpDialog
   PROTECTED:
   CLASS VAR aValues
   CLASS VAR aDefines
   CLASS VAR optionsDialog

   VAR padR
   VAR info
   VAR numItems
   VAR logWidth
   VAR refString
   VAR refEvent

   * Contained control elements
   VAR treeView
   VAR buttonMgr

   EXPORTED:
   CLASS VAR options READONLY

   CLASS METHOD initClass
   CLASS METHOD parseAppEventCH

   METHOD init
   METHOD create
   METHOD resize
   METHOD addEvent
   METHOD format
   METHOD clear
   METHOD hide
   METHOD translate
   METHOD openOptions
   METHOD updateOptions
   METHOD adjustToLog
   METHOD saveLog

   METHOD readAppEventCH
   METHOD getLogEntry

   VAR logSelected
   ASSIGN METHOD logSelected
ENDCLASS


// ---------------------------------------------------------------------------
// Initialize class object with configuration data
// ---------------------------------------------------------------------------
CLASS METHOD EventSpyDialog:initClass
   LOCAL cFile, cBinary, aData

   ::options := EventSpyOptions():restore()
   cFile     := ::options:includePath + SPY_EVENT_FILE

   #pragma Info( NOINIT )
   IF File( cFile ) .AND. ReadStream( cFile, @cBinary )
      aData       := Bin2Var( cBinary )
      ::aValues   := aData[1]
      ::aDefines  := aData[2]
   ELSE
      ::parseAppEventCH()
   ENDIF   
   #pragma Info( INIT )

RETURN self


// ---------------------------------------------------------------------------
// read APPEVENT.CH, store #defines and values in the corresponding
// class variables and save binary data to file so that it is available
// immediately when the EventSpy is started again
// ---------------------------------------------------------------------------
CLASS METHOD EventSpyDialog:parseAppEventCH
   LOCAL cFile := ::options:includePath + SPY_EVENT_FILE
   LOCAL aData[2]

   ::aValues  := {}
   ::aDefines := {}

   ReadAppEventCH( @::aDefines, @::aValues )

   aData[1]   := ::aValues  
   aData[2]   := ::aDefines 

   IF .NOT. WriteStream( cFile, Var2Bin( aData ) )
      MsgBox( "Unable to save APPEVENT.CH as binary file", "Warning" )
   ENDIF
RETURN self


// ---------------------------------------------------------------------------
// Initialize form
// ---------------------------------------------------------------------------
METHOD EventSpyDialog:init( oParent, oOwner, aPos, aSize, aPP, lVisible )

   DEFAULT oParent  TO AppDesktop(), ;
           aPos     TO {14,192}, ;
           aSize    TO {327,219}, ;
           lVisible TO .F. , ;
           aPP      TO {}

   // Saved values from previous session
   IF Valtype( ::options:pos ) == "A"
      aPos :=  ::options:pos
   ENDIF

   IF Valtype( ::options:size ) == "A"
      aSize := ::options:size
   ENDIF

   AAdd( aPP, { XBP_PP_BGCLR       , GRA_CLR_PALEGRAY } )
   AAdd( aPP, { XBP_PP_COMPOUNDNAME, FONT_HELV_SMALL  } )

   ::XbpDialog:init( oParent, oOwner, aPos, aSize, aPP, lVisible )
   ::XbpDialog:drawingArea:ColorBG := GRA_CLR_PALEGRAY
   ::XbpDialog:taskList := .T.
   ::XbpDialog:title    := "Event Spy"
   ::XbpDialog:close    := {|mp1,mp2,obj| obj:hide() }

   ::drawingArea:clipChildren := .T.
   ::treeView   := XbpTreeView():new( ::drawingArea, , {4,3*4+32}, {312,152} )
   ::buttonMgr  := GroupManager():new( ::drawingArea, , {60,24} ):create()
   ::buttonMgr:maxDistance := 24

   ::padR      := AFill( Array(7), 0  )
   ::info      := AFill( Array(7), "" )
   ::logWidth  := 0
   ::refString := ""
   ::refEvent  := { xbeP_KillDisplayFocus,,, self }
RETURN self


// ---------------------------------------------------------------------------
// Request system resources
// ---------------------------------------------------------------------------
METHOD EventSpyDialog:create( oParent, oOwner, aPos, aSize, aPP, lVisible )
   LOCAL aMinSize

   ::XbpDialog:create( oParent, oOwner, aPos, aSize, aPP, lVisible )

   ::treeView:create()

   IF ::options:useFixedFont
      ::treeview:setFontCompoundName( FONT_COURIER_SMALL )
   ELSE
      ::treeview:setFontCompoundName( FONT_HELV_SMALL    )
   ENDIF

   ::numItems := 0
   ::addEvent( xbe_None, NIL, NIL, SetAppWindow() )

   ::buttonMgr:addButton( "Hide"    , {|| ::hide()        }, .T., .T. )
   ::buttonMgr:addButton( "Clear"   , {|| ::clear()       }, .T., .T. )
   ::buttonMgr:addButton( "Save Log", {|| ::saveLog()     }, .T., .T. )
   ::buttonMgr:addButton( "Options" , {|| ::openOptions() }, .T., .T. )
   ::buttonMgr:setPos( {0,8} )

   aMinSize    := ::buttonMgr:minSize()
   aMinSize[1] += (2*4)
   aMinSize[2] += (4*4 + 32)

   aMinSize    := ::calcFrameRect( { 0, 0, aMinSize[1], aMinSize[2] } )
   ::minSize   := { aMinSize[3], aMinSize[4] }
   aSize       := ::currentSize()
   aSize[1]    := Max( aSize[1], aMinSize[1] )
   aSize[2]    := Max( aSize[2], aMinSize[2] )

   // this results in a resize event and :resize() is called in turn
   ::xbpDialog:setSize( aSize )

RETURN self


// ---------------------------------------------------------------------------
// Resize the dialog contents
// ---------------------------------------------------------------------------
METHOD EventSpyDialog:resize
   LOCAL aSize := ::drawingArea:currentSize()

   ::buttonMgr:center( aSize[1] )
   aSize:= { aSize[1]-8, aSize[2] - (4*4 + 32)  }
   ::treeview:setSize( aSize, .F. )
   ::drawingArea:invalidateRect()

   // save pos/size only if Event Spy is neither minimized nor maximized
   IF ::getFrameState() == XBPDLG_FRAMESTAT_NORMALIZED
      ::options:pos  := ::currentPos()
      ::options:size := ::currentSize()
   ENDIF

RETURN self


// ---------------------------------------------------------------------------
// Resize the dialog to the width of the log
// ---------------------------------------------------------------------------
METHOD EventSpyDialog:adjustToLog
   LOCAL aSize := ::drawingArea:currentSize()
   LOCAL aRect := ::calcFrameRect( { 0, 0, ::logWidth, aSize[2] } )

RETURN ::xbpDialog:setSize( {aRect[3], aRect[4]} )


// ---------------------------------------------------------------------------
// Add an event as string
// ---------------------------------------------------------------------------
METHOD EventSpyDialog:addEvent( nEvent, mp1, mp2, oXbp )
   LOCAL cString, i, oItem, aItem

   DO CASE
   CASE oXbp == ::treeView .OR. ;
      ( oXbp <> NIL .AND. oXbp:isDerivedFrom( "XbpTreeViewItem" ) .AND. AScan( ::treeView:rootItem:getChildItems(), oXbp ) > 0 )
      // Do not add events when the current XBP is the treeview
      // Otherwise we end up in endless recursion

         RETURN self

   CASE ( nEvent == xbeM_Motion .OR. nEvent == xbeP_Move) .AND. ;
        .NOT. ::options:showMotion

      // Filter xbeM_Motion and xbeP_Move when the option is not selected
      RETURN self

   CASE nEvent == xbeP_Paint  .AND. .NOT. ::options:showPaint
      // Filter xbeP_Paint when the option is not selected
      RETURN self

   CASE .NOT. ::options:showToSpy
      IF oXbp == self .OR. IsChild( self, oXbp )
         // Filter all events sent to the Event Spy
         RETURN self

      ELSEIF ::optionsDialog <> NIL .AND. ;
           ( oXbp == ::optionsDialog .OR. IsChild( ::optionsDialog, oXbp ) )

         // Filter all events sent to the Options window
         RETURN self
      ENDIF
   ENDCASE

   IF .NOT. ::options:addHidden  .AND. ;
    ( .NOT. ::isVisible() .OR. ::getFrameState() == XBPDLG_FRAMESTAT_MINIMIZED )

      // Do not add events when the Event Spy is not visible
      RETURN self
   ENDIF


   cString := ::format( nEvent, mp1, mp2, oXbp )

   // Add the log entry to the tree view
   // and display the last tree view item
   oItem      := ::treeView:rootItem:addItem( cString )
   ::numItems ++
   ::treeView:setData( oItem )

   // Restrict number of log entries to max. setting
   IF ::numItems > ::options:maxEvents 
      aItem  := ::treeView:rootItem:getChildItems()
      i      := 1
      DO WHILE ::numItems > ::options:maxEvents 
         ::treeView:rootItem:delItem( aItem[i] )
         ::numItems --
         i ++
      ENDDO
   ENDIF

   // Force Event Spy to top if option is selected (credit to Scott McNay)
   IF ::options:stayOnTop
      ::toFront()
   ENDIF

   // Store the current event info for later retrieval via :getLogEntry()
   // (credit to Scott McNay)
   oItem:setData( { nEvent, mp1, mp2, oXbp } )

   // Store the current event info as reference for auto adjustment of width
   IF Len( cString ) > Len( ::refString )
      ::refString := cString
      ::refEvent  := oItem:getData()
   ENDIF
RETURN self


// ---------------------------------------------------------------------------
// Return formatted string according to selected options
// ---------------------------------------------------------------------------
METHOD EventSpyDialog:format( nEvent, mp1, mp2, oXbp )
   LOCAL cString, i, nPadL

   AFill( ::info, "" )

   IF ::options:showFocus
      // credit to Scott McNay
      IF oXbp <> NIL .AND. IsMethod( oXbp, "hasInputFocus") .AND. oXbp:hasInputFocus()
         ::info[1] := "(*)"
      ELSEIF ::options:useFixedFont
         ::info[1] := "   "
      ENDIF
   ENDIF

   IF ::options:showThreadID
      ::info[2] := "T" + LTrim( Str(ThreadID()) ) 
   ENDIF

   IF ::options:showTimeStamp
      // credit to Scott McNay
      ::info[3] := Time() + ":" + Transform( (Seconds() * 10) % 100,"@L0 ## ") 
   ENDIF

   // Get the #define string representation from the numeric value
   i := AScan( ::aValues, nEvent )
   IF i > 0
      ::info[4] := ::aDefines[i]
   ELSEIF nEvent > 31 .AND. nEvent <= 255
      ::info[4] := "Keyboard in XbpCrt: " + Chr( nEvent )
   ELSEIF nEvent > xbeP_User
      ::info[4] := "xbeP_User + "+Ltrim(Str(nEvent-xbeP_User))
   ELSE
      IF nEvent > xbeB_Event
         ::info[4] := "Unknown event: xbeB_Event + "+Ltrim(Str(nEvent-xbeB_Event))
      ELSE
         ::info[4] := "Unknown event: "+Ltrim(Str(nEvent))
      ENDIF
   ENDIF

   IF ::options:showShort .AND. "xbe" $ (cString := ::info[4])
      // credit to Ken Levitt
      IF "xbeK_" $ cString
          CutStr( "xbeK_", @::info[4] )  
      ELSE
         ::info[4] := CutStr( "xbe", @cString )
         ::info[4] += CutStr( "_"  , @cString )
         ::info[4] += cString
      ENDIF
   ENDIF

   IF ::options:useFixedFont
      // pad numerics to a length of min. 3 chars to enhance
      // readability of pos/size arrays
      nPadL := 3
   ENDIF

   // Get the #define string for keyboard events
   // (the pressed key arrives in mp1)
   IF nEvent == xbeP_Keyboard
      i := AScan( ::aValues, mp1 )
      IF i > 0
         ::info[5] := ::aDefines[i]

         IF ::options:showShort .AND. ("K_" $ ::info[5])
            // credit to Ken Levitt
            CutStr( "K_", @::info[5] )
         ENDIF
      ELSEIF mp1 > 31 .AND. mp1 <= 255
         ::info[5] := Chr( mp1 )
      ENDIF
   ELSE
      ::info[5] := ToString( mp1, nPadL )
   ENDIF

   IF ::options:showMp2
      ::info[6] := ToString( mp2, nPadL )
   ENDIF

   ::info[7] := ToString( oXbp )

   // build the info string that is displayed
   cString := ""
   IF ::options:useFixedFont
      // pad info to enhance readability
      FOR i:=1 TO 6
         IF ::info[i] == "" .OR. (i==6 .AND. .NOT. ::options:showMp2 )
            LOOP
         ENDIF
         ::padR[i] := Max ( ::padR[i], Len( ::info[i] ) )
         cString   += PadR( ::info[i], ::padR[i] ) + " "
      NEXT
      // Last info is not padded
      cString += ::info[i]
   ELSE
      FOR i:=1 TO 7
         IF ::info[i] == "" .OR. (i==6 .AND. .NOT. ::options:showMp2 )
            LOOP
         ENDIF
         cString += ::info[i] + " "
      NEXT
   ENDIF

RETURN cString


// ---------------------------------------------------------------------------
// Clear the event list with :destroy()/:create() on the XbpTreeView object
// ---------------------------------------------------------------------------
METHOD EventSpyDialog:clear
   LOCAL aPos  := ::treeView:currentPos()
   LOCAL aSize := ::treeView:currentSize()

   ::treeView:destroy()
   ::treeView:create( ::drawingArea,, aPos, aSize )
   ::updateOptions()
   ::addEvent( xbe_None, NIL, NIL, SetAppWindow() )
   ::numItems := 1

   AFill( ::padR, 0  )
   AFill( ::info, "" )

   ::refString := ""
   ::refEvent  := { xbeP_KillDisplayFocus,,, self }

RETURN self


// ---------------------------------------------------------------------------
// Hide the Event Spy and save current options
// ---------------------------------------------------------------------------
METHOD EventSpyDialog:hide

   ::setFrameState( XBPDLG_FRAMESTAT_MINIMIZED )

   // :save() is a CLASS METHOD and ::options is an instance! Therefore,
   // the instance must be passed as parameter to the class method
   ::options:save( ::options )         
RETURN self


// ---------------------------------------------------------------------------
// Translate a numeric event to char and vice versa
// ---------------------------------------------------------------------------
METHOD EventSpyDialog:translate( xEvent )
   LOCAL i, xRet

   IF Valtype( xEvent ) == "C"
      i := AScan( ::aDefines, xEvent )
      IF i > 0
         xRet := ::aValues[i]
      ELSE
         xRet := -1
      ENDIF
   ELSE
      i := AScan( ::aValues , xEvent )
      IF i > 0
         xRet := ::aDefines[i]
      ELSE
         xRet := "Unknown event"
      ENDIF
   ENDIF

RETURN xRet


// ---------------------------------------------------------------------------
// Open the options dialog
// ---------------------------------------------------------------------------
METHOD EventSpyDialog:openOptions
   LOCAL aPos, aSize
   
   IF ::optionsDialog == NIL
      // open options dialog to the right of EventSpy
      aPos    := ::currentPos()
      aPos[1] += ::currentSize()[1] + 4
      ::optionsDialog := EventSpyOptionsDialog():new( AppDesktop(), self, aPos ):create()
   ENDIF

   // adjust position so that window is always entirely visible
   aPos  := ::optionsDialog:currentPos()
   aSize := ::optionsDialog:currentSize()

   aPos[1] := Max( 0, Min( aPos[1], AppDesktop():currentSize()[1]-aSize[1] ) )
   aPos[2] := Max( 0, Min( aPos[2], AppDesktop():currentSize()[2]-aSize[2] ) )

   ::optionsDialog:setPos( aPos )
   ::optionsDialog:show()

   SetAppFocus( ::optionsDialog )
RETURN self


// ---------------------------------------------------------------------------
// change the font according to ::options
//
// adjust width to log if necessary (credit to Ken Levitt)
// ---------------------------------------------------------------------------
METHOD EventSpyDialog:updateOptions
   LOCAL oPS, aBox

   IF ::options:useFixedFont
      ::treeview:setFontCompoundName( FONT_COURIER_SMALL )
   ELSE
      ::treeview:setFontCompoundName( FONT_HELV_SMALL )
   ENDIF

   IF ::options:autoSize
      // Re-format reference string for log-width calculation
      ::refString := ::format( ::refEvent[1], ::refEvent[2], ::refEvent[3], ::refEvent[4] )
      oPS         := ::treeView:lockPS()
      aBox        := GraQueryTextBox( oPS, ::refString )
      ::treeView:unlockPS( oPS )

      ::logWidth  := aBox[3,1]-aBox[2,1]
      ::adjustToLog()
   ENDIF

   AFill( ::padR, 0 )
RETURN self


// ---------------------------------------------------------------------------
// Save the event log to EVENTSPY.LOG (credit to Scott McNay)
// ---------------------------------------------------------------------------
METHOD EventSpyDialog:saveLog
   LOCAL cText := ""
   
   AEval( ::treeView:rootItem:getChildItems(), ;
          {|o| cText += o:caption + CRLF } )

   // LOG is written to current directory
   WriteStream( "EVENTSPY.LOG", cText )
RETURN self


// ---------------------------------------------------------------------------
// Method for passing Appevent.ch data to the program
// that uses the Event Spy (credit to Scott McNay)
// ---------------------------------------------------------------------------
METHOD EventSpyDialog:ReadAppeventCH( aDefines, aValues )

   // Just in case the params are passed by reference
   IF Valtype( aDefines ) <> "A"
      aDefines := {}
   ENDIF

   IF Valtype( aValues ) <> "A"
      aValues := {}
   ENDIF

   ASize( aDefines  , Len( ::aDefines ) )
   ASize( aValues   , Len( ::aValues  ) )

   ACopy( ::aDefines, aDefines )
   ACopy( ::aValues , aValues  )
   
RETURN self


// ---------------------------------------------------------------------------
// Methods for looking up raw data of the event log from another program
// that uses the Event Spy (credit to Scott McNay)
//
// NOTE: the method :logSelected() is an ASSIGN method
//       so that the :itemSelected code block for the
//       :treeView can be set or voided when
//                  EventSpy():logSelected := bBlock | NIL
//       occurs in an external program
// ---------------------------------------------------------------------------
METHOD EventSpyDialog:logSelected( bBlock )
   IF Valtype( bBlock ) == "B"
      ::logSelected := bBlock
      ::treeView:itemSelected := {|oItem| Eval( ::logSelected, oItem:getData() ) }
   ELSE
      ::logSelected := NIL
      ::treeView:itemSelected := NIL
   ENDIF
RETURN ::logSelected


METHOD EventSpyDialog:getLogEntry( nPos )
   LOCAL aItems   := ::treeview:rootItem:getChildItems()
   LOCAL aRawData[4]

   IF Valtype( nPos ) <> "N"
      nPos := Len( aItems )
   ENDIF

   IF nPos > 0 .AND. nPos <= Len( aItems )
      aRawData := aItems[nPos]:getData()
   ENDIF
RETURN aRawData


******************************************************************************
* Class for managing the position of grouped Xbase Parts in
* sizeable dialog windows. All XBPs of one group have the same size!
*
* This is adopted from the Alaska Software Xbase++ example program
*
*   ..\samples\solution\resize\groupmgr.prg
*
******************************************************************************
STATIC CLASS GroupManager FROM XbpPartHandler
   PROTECTED:
   VAR aXbpSize         // size of single Part
   VAR aMinSize         // min size of all Parts
   VAR aXbps            // array with XbpParts
   METHOD centerH       // center Parts horizontally
   METHOD centerV       // center Parts vertically

   EXPORTED:
   VAR maxDistance      // max. distance between Parts
   VAR currentDistance  // current distance between Parts
   VAR horizontal       // .T. => Parts are displayed horizontally
   
   METHOD init
   METHOD addButton     // creates an XbpPushbutton
   METHOD addCheckbox   // creates an XbpCheckbox
   METHOD part          // returns an Xbase Part by ordinal position
   METHOD setPos        // reposition parts group to given position
   METHOD center        // centers all Parts within the group
   METHOD minSize       // returns minimum size of the Part group
   METHOD currentPos    // returns current pos of the Part group
   METHOD currentSize   // returns current size of the Part group
ENDCLASS


// ---------------------------------------------------------------------------
// Initialize object 
// ---------------------------------------------------------------------------
METHOD GroupManager:init( oParent, oOwner, aSize )

   DEFAULT oParent TO SetAppWindow(), ;
           oOwner  TO oParent

   ::xbpPartHandler:init( oParent, oOwner )
   ::aXbps           := {}
   ::aXbpSize        := AClone( aSize )

   ::currentDistance := 0
   ::maxDistance     := 24
   ::horizontal      := .T.

RETURN self


// ---------------------------------------------------------------------------
// Creates a new pushbutton and adds it to the protected array 
// ---------------------------------------------------------------------------
METHOD GroupManager:addButton( xCaption, bActivate, lTabstop, lPointerFocus, nGroup )
   LOCAL oButton

   DEFAULT lTabStop      TO .T. , ;
           lPointerFocus TO .T. , ;
           nGroup        TO XBP_NO_GROUP

   oButton              := XbpPushbutton():new( ::setParent(), ::setOwner(), ;
                                                NIL          , ::aXbpSize )
   oButton:caption      := xCaption
   oButton:activate     := bActivate
   oButton:tabStop      := lTabStop
   oButton:pointerFocus := lPointerFocus
   oButton:create()

   AAdd( ::aXbps, oButton )

   IF ::aMinSize == NIL
      ::aMinSize := Aclone( ::aXbpSize )
   ELSEIF ::horizontal
      ::aMinSize[1] += ::aXbpSize[1]
   ELSE
      ::aMinSize[2] += ::aXbpSize[2]
   ENDIF

RETURN self



// ---------------------------------------------------------------------------
// Creates a new Checkbox and adds it to the protected array 
// ---------------------------------------------------------------------------
METHOD GroupManager:addCheckbox( xCaption, bSelected, lTabstop, lPointerFocus, nGroup )
   LOCAL oCheckbox

   DEFAULT lTabStop      TO .T. , ;
           lPointerFocus TO .T. , ;
           nGroup        TO XBP_NO_GROUP

   oCheckbox              := XbpCheckbox():new( ::setParent(), ::setOwner(), ;
                                                NIL          , ::aXbpSize )
   oCheckbox:caption      := xCaption
   oCheckbox:Selected     := bSelected
   oCheckbox:tabStop      := lTabStop
   oCheckbox:pointerFocus := lPointerFocus
   oCheckbox:create()

   AAdd( ::aXbps, oCheckbox )

   IF ::aMinSize == NIL
      ::aMinSize := Aclone( ::aXbpSize )
   ELSEIF ::horizontal
      ::aMinSize[1] += ::aXbpSize[1]
   ELSE
      ::aMinSize[2] += ::aXbpSize[2]
   ENDIF

RETURN self


// ---------------------------------------------------------------------------
// Returns a Part by ordinal position 
// ---------------------------------------------------------------------------
METHOD GroupManager:part( nIndex )
RETURN ::aXbps[ nIndex ]


// ---------------------------------------------------------------------------
// Sets the position for the "anchor" Part and repositions all other 
// Parts from there 
// ---------------------------------------------------------------------------
METHOD GroupManager:setPos( aPos, lPaint )
   LOCAL aXbpPos, i, nCount := Len( ::aXbps )

   DEFAULT lPaint TO .T.

   IF ::horizontal                     // horizontal positioning left->right
      i := 1
   ELSE                                // vertical positioning bottom->top
      i := nCount
   ENDIF
   aXbpPos := AClone( aPos )
   ::aXbps[i]:setPos( aXbpPos )

   IF ::horizontal                     // position buttons from left to right
      FOR i := 2 TO nCount
         aXbpPos[1] += ::currentDistance + ::aXbpSize[1]
         ::aXbps[i]:setPos( aXbpPos, lPaint )
      NEXT 
   ELSE                                // position buttons from bottom to top
      nCount --
      FOR i := nCount TO 1 STEP -1
         aXbpPos[2] += ::currentDistance + ::aXbpSize[2]
         ::aXbps[i]:setPos( aXbpPos, lPaint )
      NEXT 
   ENDIF

RETURN self


// ---------------------------------------------------------------------------
// Centers the Part group within a given width and adds an offset 
// ---------------------------------------------------------------------------
METHOD GroupManager:center( nWidth, nOffset )
   IF ::horizontal
      ::centerH( nWidth, nOffset )
   ELSE
      ::centerV( nWidth, nOffset )
   ENDIF
RETURN self


// ---------------------------------------------------------------------------
// Centers Xbase Parts horizontally with variable distance between parts
// ---------------------------------------------------------------------------
METHOD GroupManager:centerH( nWidth, nLeftOffset )
   LOCAL aPos  := ::aXbps[1]:currentPos()
   LOCAL nCount:= Len( ::aXbps )
   LOCAL nDist := nWidth - ( nCount * ::aXbpSize[1] )
   LOCAL i

   DEFAULT nLeftOffset TO 0
   nDist -= nLeftOffSet

   // Variable distance between Parts
   IF nCount > 1
      ::currentDistance := Min( ::maxDistance, Max( 0, nDist / nCount ) )
   ELSE
      ::currentDistance := 0
   ENDIF

   aPos[1] := ( nWidth - ( nCount * ::aXbpSize[1] ) - ;
                         ( ( nCount - 1 ) * ::currentDistance ) ) / 2
   aPos[1] += nLeftOffset
   aPos[1] := Max( aPos[1], nLeftOffset )

   ::aXbps[1]:setPos( aPos )
   FOR i:=2 TO nCount
      aPos[1] += ::currentDistance + ::aXbpSize[1]
      ::aXbps[i]:setPos( aPos )
      // force repaint in case XBP overlap
      ::aXbps[i]:invalidateRect()
   NEXT   

RETURN self


// ---------------------------------------------------------------------------
// Centers Xbase Parts vertically with variable distance between parts 
// ---------------------------------------------------------------------------
METHOD GroupManager:centerV( nHeight, nBottomOffset )
   LOCAL aPos  := ::aXbps[1]:currentPos()
   LOCAL nCount:= Len( ::aXbps )
   LOCAL nDist := nHeight - ( nCount * ::aXbpSize[2] )
   LOCAL i

   DEFAULT nBottomOffset TO 0
   nDist -= nBottomOffSet

   // Variable distance between Parts
   IF nCount > 1
      ::currentDistance := Min( ::maxDistance, Max( 0, nDist / nCount ) )
   ELSE
      ::currentDistance := 0
   ENDIF

   aPos[2] := ( nHeight - ( nCount * ::aXbpSize[2] ) - ;
                          ( ( nCount - 1 ) * ::currentDistance ) ) / 2
   aPos[2] += nBottomOffset
   aPos[2] := Max( aPos[2], nBottomOffset )

   ATail(::aXbps):setPos( aPos )
   nCount --

   FOR i:= nCount TO 1 STEP -1
      aPos[2] += ::currentDistance + ::aXbpSize[2]
      ::aXbps[i]:setPos( aPos, .F. )
   NEXT   

RETURN self


// ---------------------------------------------------------------------------
// Position of lower left Part
// ---------------------------------------------------------------------------
METHOD GroupManager:currentPos
   LOCAL aPos

   IF ::horizontal
      aPos := ::aXbps[1]:currentPos()
   ELSE
      aPos := Atail(::aXbps):currentPos()
   ENDIF
RETURN aPos


// ---------------------------------------------------------------------------
// Returns the area that is occupied by all Parts 
// ---------------------------------------------------------------------------
METHOD GroupManager:currentSize
   LOCAL aPos, aSize
   
   IF Len( ::aXbps ) == 1
      RETURN AClone( ::aXbpSize )      // ** RETURN **

   ELSEIF ::horizontal
      aPos  := ::aXbps[1]:currentPos()
      aSize := ATail( ::aXbps ):currentPos()

   ELSE
      aPos  := ATail( ::aXbps ):currentPos()
      aSize := ::aXbps[1]:currentPos()

   ENDIF    

   aSize[1] += ::aXbpSize[1]
   aSize[2] += ::aXbpSize[2]
   aSize[1] -= aPos[1]
   aSize[2] -= aPos[2]
RETURN aSize


// ---------------------------------------------------------------------------
// The minimum area required to display all Parts 
// ---------------------------------------------------------------------------
METHOD GroupManager:minSize
RETURN AClone( ::aMinSize )



/*
 *
 * Some utility functions required for the Event Spy
 *
 */

******************************************************************************
* Read APPEVENT.CH 
* Fill #defines and their numeric values into arrays
*
******************************************************************************
STATIC PROCEDURE ReadAppeventCH( aDefine, aValue )
   LOCAL cPath, cBuffer, cName, cValue, cText , i

   // Find APPEVENT.CH from SET INCLUDE= path
   cPath := GetEnvDir( "INCLUDE", "APPEVENT.CH" )

   IF Empty( cPath ) .AND. .NOT. File( ".\AppEvent.ch" )
      MsgBox( "Error: APPEVENT.CH not found" + Chr(13) + ;
              "Check your SET INCLUDE= environment variable!" )
      QUIT
   ENDIF

   // Read the file and remove comment lines at the beginning
   cText := CRLF + StrTran( MemoRead( cPath + "APPEVENT.CH" ), Chr(9), " " )

   cBuffer := ""
   DO WHILE CRLF+"//" $ cText
      cBuffer += CutStr( CRLF +"//", @cText ) + CRLF
      CutStr( CRLF, @cText )
      IF CRLF+"//" $ cBuffer
         cText := cBuffer + cText
         cBuffer := ""
      ENDIF
   ENDDO
   cText := cBuffer + cText
   
   // Remove OS-specific constants
   #ifdef __WIN32__
      cName := "__OS2__"
   #else
      cName := "__WIN32__"
   #endif

   DO WHILE cName $ cText
      cBuffer := CutStr( cName, @cText )
      cBuffer := RCutStr( "#ifdef", cBuffer )
      CutStr( CRLF, @cText )
      CutStr( "#endif", @cText )
      CutStr( CRLF, @cText )
      cText := cBuffer + cText
   ENDDO

   // Read all lines that contain #define
   CutStr( "#define", @cText )

   DO WHILE ! Empty( cText )
      cText  := LTrim( cText )
      cName  := CutStr( " ", @cText )
      cValue := AllTrim( CutStr( CRLF, @cText ) )

      IF "//" $ cValue
         cValue := Alltrim( CutStr( "//", cValue ) )
      ENDIF

      DO WHILE (i := AScan( aDefine, {|c| c $ cValue } ) ) > 0
         // translate stuff like (1 + xbeB_Event)
         cValue := StrTran( cValue, aDefine[i], LTrim(aValue[i]) )
      ENDDO

      IF Empty( cValue )
         // #define _APPEVENT_CH is no event
         LOOP
      ENDIF

      AAdd( aDefine, cName )
      AAdd( aValue , Padl(&cValue,10) )
      CutStr( "#define", @cText )
   ENDDO

   AEval( aValue, {|x| x:=Val(x)},,, .T. )
RETURN



******************************************************************************
* Transforms a value to a string
******************************************************************************
STATIC FUNCTION ToString( xValue, nPad )
   LOCAL imax, cType := Valtype( xValue ), cTemp

   DO CASE
   CASE cType == "A"
      imax := Len( xValue )

      IF imax == 2
         xValue := "{" + ToString(xValue[1], nPad)+","+ToString(xValue[2], nPad) + "}"
      ELSEIF imax == 4
         xValue := "{" + ToString(xValue[1], nPad)+","+ToString(xValue[2], nPad) + "," + ;
                         ToString(xValue[3], nPad)+","+ToString(xValue[4], nPad) + "}"
      ELSE
         xValue := "Array("+Ltrim(Str(imax))+")"
      ENDIF

   CASE cType == "C" 
      xValue := '"' + xValue + '"'

   CASE cType == "O" 
      cTemp := ""

      IF xValue:IsDerivedFrom( "XbpIWindow" )
         IF xValue == AppDesktop()
            cTemp := '"AppDesktop"'
         ELSE
            cTemp := 'in ' + ToString( xValue:setParent():getTitle() )
         ENDIF

      ELSEIF IsMemberVar( xValue, "caption" )
         cTemp := ToString( xValue:caption )


      ELSEIF IsMethod( xValue, "getTitle" )
         cTemp := ToString( xValue:getTitle() )

      ENDIF

      xValue := xValue:className() + " " + cTemp

   CASE cType == "L" ; xValue := IIf( xValue, "TRUE","FALSE" )
   CASE cType == "N" 
      xValue := Ltrim( Str( xValue ) )
      IF Valtype(nPad) == "N" .AND. Len( xValue ) < nPad
         xValue := PadL( xValue, nPad )
      ENDIF

   CASE cType == "D" ; xValue := DtoC( xValue )
   CASE cType == "U" ; xValue := "NIL"
   CASE cType == "B" ; xValue := "Code block"
   ENDCASE
RETURN xValue



******************************************************************************
* Two of the most useful functions for parsing strings
******************************************************************************
FUNCTION CutStr( cCut, cString )
   LOCAL cLeftPart, i := At( cCut, cString )

   IF i > 0
      cLeftPart := Left( cString, i-1 )
      cString   := SubStr( cString, i+Len(cCut) )
   ELSE
      cLeftPart := cString
      cString   := ""
   ENDIF
RETURN cLeftPart


FUNCTION RCutStr( cCut, cString )
   LOCAL cLeftPart, i := RAt( cCut, cString )

   IF i > 0
      cLeftPart := Left( cString, i-1 )
      cString   := SubStr( cString, i+Len(cCut) )
   ELSE
      cLeftPart := cString
      cString   := ""
   ENDIF
RETURN cLeftPart


******************************************************************************
* Read/write streams of data from/to file
*
* NOTE: cStream must be passed by reference
******************************************************************************
FUNCTION ReadStream( cFile, cStream )
   LOCAL nHandle := FOpen( cFile )
   LOCAL nFSize

   IF FError() <> 0
      RETURN .F.
   ENDIF

   nFSize  := FSeek( nHandle, 0, FS_END )
   cStream := Space( nFSize )
   FSeek( nHandle, 0, FS_SET )
   FRead( nHandle, @cStream, nFSize )
   FClose( nHandle )
RETURN ( FError() == 0 .AND. .NOT. Empty( cStream ) )



FUNCTION WriteStream( cFile, cStream )
   LOCAL nHandle := FCreate( cFile )

   IF FError() <> 0
      RETURN .F.
   ENDIF

   FWrite( nHandle, cStream, Len(cStream) )
   FClose( nHandle )
RETURN ( FError() == 0 )



******************************************************************************
* Find a directory from a GetEnv() variable that contains a particular file
* If the file is found in the current directory, this directory is returned
******************************************************************************
FUNCTION GetEnvDir( cEnv, cFile )
   LOCAL cPath := GetEnv( cEnv )
   LOCAL cDir  := ""
   LOCAL cTemp

   IF File( ".\" + cFile )
      RETURN ".\"
   ENDIF

   DO WHILE ! Empty( cPath )
      cTemp := CutStr( ";", @cPath )
      cTemp := StrTran( cTemp + "\", "\\", "\" )
      IF File( cTemp + cFile )
         cDir := cTemp
         EXIT
      ENDIF
   ENDDO
RETURN cDir


******************************************************************************
* Find out if an Xbase Part is stored in the parent-child hierarchy
******************************************************************************
FUNCTION IsChild( oParent, oChild )
   LOCAL aArray   := oParent:childList()
   LOCAL lIsChild := .F.
   LOCAL i, imax  := Len( aArray ) 

   FOR i:=1 TO imax
      IF oChild == aArray[i] .OR. IsChild( aArray[i], oChild )  
         lIsChild := .T.
         EXIT
      ENDIF
   NEXT
RETURN lIsChild
