//////////////////////////////////////////////////////////////////////
//
//  ERRORSYS.PRG
//
//  Copyright:
//      Alaska Software Inc., (c) 1997-2000. All rights reserved.
//
//  Contents:
//      Install default error code block
//
//  Remarks:
//      Function ErrorSys() is always called at program startup
//
//////////////////////////////////////////////////////////////////////

#include 'dcdialog.ch'
#include 'error.ch'

***********************************
* Name of error log without extension
***********************************
#define EHS_ERRORLOG "XPPERROR"

/*
 *  language specific string constant which are used in the error handler
 */
#define EHS_CANCEL            "Cancel"
#define EHS_EXIT_WITH_LOG     "Exit with LOG file"
#define EHS_RETRY             "Retry"
#define EHS_IGNORE            "Ignore"
#define EHS_OS_ERROR          ";Operating system error : "
#define EHS_CALLED_FROM       "Called from"
#define EHS_XPP_ERROR_MESSAGE "Xbase++ Error Message"
#define EHS_ERROR             "Error "
#define EHS_WARNING           "Warning "
#define EHS_DESCRIPTION       ";Description : "
#define EHS_FILE              ";File : "
#define EHS_OPERATION         ";Operation : "
#define EHS_LOG_OPEN_FAILED   "Unable to open error log file"
#define EHS_ERROR_LOG_OF      "ERROR LOG of "
#define EHS_DATE              " Date:"
#define EHS_XPP_VERSION       "Xbase++ version     :"
#define EHS_OS_VERSION        "Operating system    :"
#define EHS_LOG_WRITTEN_TO(cFile) "Error log was written to the file "+ cFile

***********************************
* Install default error code block
***********************************
PROCEDURE ErrorSys()
   ErrorBlock( {|o| StandardEH(o)} )
RETURN



*************************************
* Default error handler function
*************************************
STATIC FUNCTION StandardEH( oError )
   LOCAL i, cMessage, aOptions, nOption, nSeverity
   LOCAL row, col

   /* Check if error is handled automatically */
   DO CASE

   /* Division by zero results in 0 */
   CASE oError:genCode == XPP_ERR_ZERODIV
      RETURN 0

   /* Zero divide by zero is also 0 and may occur
    * in operations like /, /=, % and %=
    */
   CASE oError:genCode == XPP_ERR_NUMERR
      IF "/" $ oError:operation .OR. "%" $ oError:operation
          IF oError:args[1] = 0
             IF len(oError:args) = 1
                 RETURN 0
             ELSEIF oError:args[2] = 0
                 RETURN 0
             ENDIF
          ENDIF
      ENDIF

   /* Error opening a file on a network */
   CASE oError:genCode == XPP_ERR_OPEN  .AND. ;
        oError:osCode  == 32            .AND. ;
        oError:canDefault
      RETURN(.F.)

   /* No lock is set */
   CASE oError:genCode == XPP_ERR_APPENDLOCK .AND. ;
        oError:canDefault
      RETURN(.F.)

   ENDCASE


   /* No default handling defined: create error message */
   cMessage := ErrorMessage( oError )

   /* Array for selection */
#ifdef DEBUG
   aOptions := { EHS_CANCEL, EHS_EXIT_WITH_LOG }
#else
   aOptions := { EHS_CANCEL }
#endif

   IF oError:canRetry
      AAdd( aOptions, EHS_RETRY )
   ENDIF

   IF oError:canDefault
      AAdd( aOptions, EHS_IGNORE )
   ENDIF

   IF ! Empty( oError:osCode )
      cMessage += EHS_OS_ERROR + LTrim(Str(oError:osCode))
   ENDIF

   IF AppType() <> APPTYPE_PM

      /* Display Alert() Box possible ? */
      IF SetAppWindow() != NIL
         i   := 0
         row := Row()
         col := Col()
         DO WHILE i == 0
            i := Alert( cMessage, aOptions )
         ENDDO
         SetPos( row, col )

         /* Perform selected option */
         IF ! Empty( i )
       DO CASE
       CASE aOptions[i] == EHS_IGNORE
          RETURN .F.
       CASE aOptions[i] == EHS_RETRY
          RETURN .T.
       CASE aOptions[i] == EHS_CANCEL
          Break( oError )
       CASE aOptions[i] == EHS_EXIT_WITH_LOG
          ErrorLog( oError, 2 )
       ENDCASE
    ENDIF
      ELSE
         /* There is no console window */
#ifdef DEBUG
    ErrorLog( oError, 2 )
#endif
      ENDIF

      /*
       * Program can not or should not be continued
       * Set error level and terminate program !
       */

      ErrorLevel(1)
      QUIT
   ENDIF

   IF oError:canDefault .AND. oError:canRetry
      nOption := XBPMB_ABORTRETRYIGNORE
   ELSEIF oError:canRetry
      nOption := XBPMB_RETRYCANCEL
   ELSEIF oError:canDefault
      nOption := XBPMB_OKCANCEL
   ELSE
      nOption := XBPMB_CANCEL
   ENDIF

   i := 1
   DO WHILE ! Empty( ProcName(++i) )
      cMessage+= ";" + EHS_CALLED_FROM +" " + Trim( ProcName(i) )   + "(" + ;
                             LTrim( Str( ProcLine(i) ) ) + ")"
   ENDDO

   i := 0
   /* select icon for ConfirmBox() */
   DO CASE
      CASE oError:severity == XPP_ES_FATAL
           nSeverity := XBPMB_CRITICAL
      CASE oError:severity == XPP_ES_ERROR
           nSeverity := XBPMB_CRITICAL
      CASE oError:severity == XPP_ES_WARNING
           nSeverity := XBPMB_WARNING
      OTHERWISE
           nSeverity := XBPMB_INFORMATION
   ENDCASE
   /* Display ConfirmBox() */
   i := ConfirmBox( , StrTran( cMessage, ";", Chr(13) ), ;
                    EHS_XPP_ERROR_MESSAGE , ;
                    nOption , ;
                    nSeverity + XBPMB_APPMODAL+XBPMB_MOVEABLE )

   DO CASE
   CASE i == XBPMB_RET_RETRY
      RETURN (.T.)
   CASE i == XBPMB_RET_IGNORE
      RETURN (.F.)
   CASE i == XBPMB_RET_CANCEL
#ifdef DEBUG
         IF ConfirmBox(, EHS_EXIT_WITH_LOG, EHS_XPP_ERROR_MESSAGE, XBPMB_YESNO,;
                  XBPMB_WARNING+XBPMB_APPMODAL+XBPMB_MOVEABLE ) != XBPMB_RET_YES
              Break( oError )
         ENDIF
#else
         Break( oError )
#endif
   ENDCASE

BEGIN SEQUENCE

#ifdef DEBUG
   ErrorLog( oError, 2 )
#endif

   /* Set error level and terminate program ! */
   ErrorLevel(1)

END SEQUENCE

  QUIT

RETURN .F. /* The compiler expects a return value */


***************************************
STATIC FUNCTION ErrorMessage( oError )
*
*  Creates a string with the important Informations
*  from the error object
***************************************

   /* Check if this is an error or warning message */
   LOCAL cMessage := ;
         IIf( oError:severity > XPP_ES_WARNING, ;
                          EHS_ERROR, EHS_WARNING )

   /* Add name of subsystem or 'unkown subsytem' */
   IF Valtype( oError:subSystem ) == "C"
      cMessage += oError:subSystem
   ELSE
      cMessage += "????"
   ENDIF

   /* Add error code of subsystem */
   IF Valtype( oError:subCode ) == "N"
      cMessage += "/"+ LTrim(Str(oError:subCode))
   ELSE
      cMessage += "/????"
   ENDIF

   /* Optional: Add error description */
   IF Valtype( oError:description ) == "C"
      cMessage += EHS_DESCRIPTION + ;
                   oError:description
   ENDIF

   /* Optional: Add name of the file which were the error occured */
   IF ! Empty( oError:fileName )
      cMessage += EHS_FILE + oError:fileName
   ENDIF

   /* Optional: Add name of the operation which caused the error */
   IF ! Empty( oError:operation )
      cMessage += EHS_OPERATION + oError:operation
   ENDIF

   /* Add Thread ID of the thread on which the error occured */
   cMessage += ";Thread ID : " + ;
                LTrim(Str(oError:thread))

RETURN cMessage



************************************************
STATIC PROCEDURE ErrorLog( oError, nStackStart )
*
*  Creates a string with the important Informations
*  from the error object
************************************************
   LOCAL i:=0, bError := ErrorBlock( {|e| Break(e)} )
   LOCAL cErrorLog
   LOCAL cExtension:= "LOG"
   LOCAL lPrint, lConsole, cAlternate, lAlternate, lExtra
   local narea, cDescription, oTracker, GetList[0], cErrorDump

   /* Save current printer related settings, turn printer off */

   lPrint     := Set( _SET_PRINTER )
   lConsole   := Set( _SET_CONSOLE )
   cAlternate := Set( _SET_ALTFILE )
   lAlternate := Set( _SET_ALTERNATE )
   lExtra     := Set( _SET_EXTRA, .F. )

   SET PRINTER OFF
   SET CONSOLE ON

   IF SetAppWindow() == NIL ;
     .OR. .NOT. SetAppWindow():isDerivedFrom( RootCrt() )
      SET CONSOLE OFF
   ENDIF

   /* Trap errors that might occur while opening the ALTERNATE file */
   DO WHILE .T.
      cErrorLog := 'TEMPERR' + "." + cExtension
      BEGIN SEQUENCE
        Set(_SET_ALTFILE, cErrorLog, .f.)
        Set(_SET_ALTERNATE, .t.)
      RECOVER
        /*
         * ALTERNATE file could not be opened:
         * try other filename
         */
        cExtension := PadL(++i,3,"0")
        IF i > 999
           IF AppType() <> APPTYPE_PM
              TONE(660,5)
              ? EHS_LOG_OPEN_FAILED
              IF !Set( _SET_CONSOLE )
                  OutErr( CHR(10)+CHR(13) + EHS_LOG_OPEN_FAILED )
              ENDIF
           ELSE
              MsgBox( EHS_LOG_OPEN_FAILED )
           ENDIF
           ErrorLevel(1)
           QUIT
        ENDIF
        LOOP
      END SEQUENCE
      EXIT
   ENDDO

   ? Replicate( "=", 78 )
   ? EHS_ERROR_LOG_OF +'"'+ appName(.T.) +'"'+ EHS_DATE, DC_XtoC(date()), time()
   ? Replicate( "-", 78 )
   ? EHS_XPP_VERSION        , Version()+"."+Version(3)
   ? "eXPress++ version   :", DC_VersionExpress()+"."+DC_VersionExpress(3)
   ? EHS_OS_VERSION  , Os()
   ? "User                :", Set_User()
   ? "Database path       :", set(_SET_PATH)
   ? Replicate( "-", 78 )
   ? "ERROR OBJECT:"
   ? Replicate( "-", 78 )
   ? "oError:args         :"
   IF Valtype(oError:Args)=="A"
      AEval( oError:Args, ;
             {|x,y| Qout( Space(9),"-> VALTYPE:", y:=Valtype(x) )  , ;
                     IIf( y=="O", QQout( " CLASS:", x:className() ), ;
                                  QQout( " VALUE:", Var2Char(x) ) ) } )
   ELSE
      Qout( Space(10),"-> NIL" )
   ENDIF

   ? "oError:canDefault   :" , oError:canDefault
   ? "oError:canRetry     :" , oError:canRetry
   ? "oError:canSubstitute:" , oError:canSubstitute
   ? "oError:cargo        :" , oError:cargo
   ? "oError:description  :" , oError:description
   ? "oError:filename     :" , oError:filename
   ? "oError:genCode      :" , oError:genCode
   ? "oError:operation    :" , oError:operation
   ? "oError:osCode       :" , oError:osCode
   ? "oError:severity     :" , oError:severity
   ? "oError:subCode      :" , oError:subCode
   ? "oError:subSystem    :" , oError:subSystem
   ? "oError:thread       :" , oError:thread
   ? "oError:tries        :" , oError:tries

   ? Replicate( "-", 78 )
   ? "CALLSTACK:"
   ? Replicate( "-", 78 )

   i := nStackStart
   DO WHILE ! Empty( ProcName(++i) )
      ? EHS_CALLED_FROM, Trim( ProcName(i) )   + "(" + ;
                 LTrim( Str( ProcLine(i) ) ) + ")"
   ENDDO

   //------------------------------------------------------------------
   // Workarea status
   //------------------------------------------------------------------
   ? replicate( "-", 78 )
   ? "WORKAREA STATUS:"
   ? replicate( "-", 78 )
   ? " Area Alias         Record     Index        Order      Eof Bof Del Found"
   ? replicate( "-", 78 )
   //  999--XXXXXXXXXX-999999/999999-XXXXXXXXXXXX-XXXXXXXXXX-XXX-XXX-XXX-XXX
   narea := select()
   for i := 1 to 255
      if !empty(alias((i)))
         dbselectarea(i)
         ? iif(i==narea, "*", " ") + str(i,3) + "  " + ;
           padr(alias(),10) + " " + ;
           str(recno(),6) + "/" + padr(ltrim(str(lastrec(),6)),6) + " " + ;
           padr(ordbagname(),12) + " " + ;
           padr(ordsetfocus(),10) + " " + ;
           iif(eof(),     "Yes", "No ") + " " + ;
           iif(bof(),     "Yes", "No ") + " " + ;
           iif(deleted(), "Yes", "No ") + " " + ;
           iif(found(),   "Yes", "No ")
      endif
   next i
   ?

   SET ALTERNATE TO
   SET ALTERNATE OFF

//   IF AppType() <> APPTYPE_PM
//      TONE(660,5)
//      ? EHS_LOG_WRITTEN_TO(cErrorLog)
//      IF !Set( _SET_CONSOLE )
//          OutErr( CHR(10)+CHR(13) + EHS_LOG_WRITTEN_TO(cErrorLog) )
//      ENDIF
//   ELSE
//      MsgBox( EHS_LOG_WRITTEN_TO(cErrorLog)  )
//   ENDIF

   /* Restore previous settings */
   Set( _SET_PRINTER,   lPrint)
   Set( _SET_CONSOLE,   lConsole)
   Set( _SET_ALTFILE,   cAlternate)
   Set( _SET_ALTERNATE, lAlternate)
   Set( _SET_EXTRA,     lExtra)

cErrorDump := MemoRead(cErrorLog)
MemoWrit(EHS_ERRORLOG+'.LOG',MemoRead(EHS_ERRORLOG+'.LOG')+cErrorDump)

cDescription := ''
@ 0,0 DCSAY 'Please tell us what you were doing when this Error occurred!' ;
   SAYSIZE 0

@ 1,0 DCMULTILINE cDescription ;
   FONT          "10.Courier" ;
   NOHORIZSCROLL ;
   SIZE          80,10

DCREAD GUI ;
   FIT ;
   ADDBUTTONS ;
   MODAL

oTracker := Tracker():new()
oTracker:priority := 'A'
oTracker:category := 'Defect'
oTracker:short := 'Runtime Error in Test System'
oTracker:description := cDescription
oTracker:errordump := cErrorDump

oTracker:addRecord()

ErrorBlock(bError)

RETURN
